// C++ code by Gergely Acs <acs@crysys.hu>

#ifndef __PIN_SKETCH_H
#define __PIN_SKETCH_H

#include <NTL/vec_ZZ.h>
#include <NTL/GF2XFactoring.h>
#include <NTL/GF2EXFactoring.h>

class PinSketch
{
    public:
	PinSketch(long p_t, long p_m )
	{
	    d = 2 * p_t + 1;
	    m = p_m;
	    BuildSparseIrred(irrP, m); // fix irreducible poly of deg m over GF(2)
	    NTL::GF2E::init(irrP); // fix field as GF(2^m)
	}

	void NumToBinElem(NTL::GF2E& e, const NTL::ZZ & a)
	{
	    NumToBinElem(e, a, m);
	}

	void GetSS(NTL::vec_GF2E& set, std::ostream& out) 
	{
	    NTL::vec_GF2E ss;
	    GetSS(set, ss);
	    OutputSS(out, ss, d, irrP);
	}	

	void GetSS(NTL::vec_GF2E& set, NTL::vec_GF2E& outSS) 
	{
	    BCHSyndromeCompute(outSS, set, d);
	}	

	long GetDiffFromSketches(std::istream& sketch1, std::istream& sketch2, std::ostream& out)
	{
	    NTL::vec_GF2E setDifference;
	    NTL::vec_GF2E ss1, ss2; // sketches of the sets

	    ReadSS(ss1, sketch1, d);
	    ReadSS(ss2, sketch2, d);

	    if (BCHSyndromeDecode(setDifference, ss1 - ss2, d))
	    {
		OutputSetDifference(out, setDifference);
		return setDifference.length();
	    }
	    
	    return -1;
	}

	long GetDiff(std::istream& ss1In, NTL::vec_GF2E& ss2In, std::ostream& out)
	{
	    NTL::vec_GF2E setDifference;
	    NTL::vec_GF2E ss1; // sketch of the first set

	    ReadSS(ss1, ss1In, d);

	    if (BCHSyndromeDecode(setDifference, ss1 - ss2In, d))
	    {
		OutputSetDifference(out, setDifference);
		return setDifference.length();
	    }
	    
	    return -1;
	}


    private:

        NTL::GF2X irrP;

	long d; // minimum distance of the code;
		// can handle set difference up to t=(d-1)/2 elements
		// sketch is (d-1)/2 elements long

	long m; // elements of the set and of the sketch are m-bit values

	void OutputSS(std::ostream &out, const NTL::vec_GF2E & syn, long d, const NTL::GF2X & irrP);

	void ReadSS(NTL::vec_GF2E & ss, std::istream &infile, long d);

	void OutputSetDifference(std::ostream &outfile, const NTL::vec_GF2E & setDifference);

	void BCHSyndromeCompute(NTL::vec_GF2E &answer, const NTL::vec_GF2E & set, long d);

	bool BCHSyndromeDecode(NTL::vec_GF2E &answer, const NTL::vec_GF2E & syndrome, long d);

	void NumToBinElem(NTL::GF2E & e, const NTL::ZZ & a, long m);
};


#endif
